@@ -1,8 +1,24 @@
+# -*- coding: utf-8 -*-
+
 """Pretty-print tabular data."""
 
 from __future__ import print_function
-
+from __future__ import unicode_literals
 from collections import namedtuple
+from platform import python_version_tuple
+
+
+if python_version_tuple()[0] < "3":
+    _int_type = int
+    _float_type = float
+    _text_type = unicode
+    _binary_type = str
+else:
+    from functools import reduce
+    _int_type = int
+    _float_type = float
+    _text_type = str
+    _binary_type = bytes
 
 
 __all__ = ["tabulate"]
@@ -16,12 +32,13 @@ DataRow = namedtuple("DataRow", ["begin", "sep", "end"])
 
 TableFormat = namedtuple("TableFormat", ["lineabove", "linebelowheader",
                                          "linebetweenrows", "linebelow",
-                                         "datarow", "usecolons",
+                                         "datarow", "padding", "usecolons",
                                          "with_header_hide",
                                          "without_header_hide"])
 
 
-_format_defaults = {"usecolons": False,
+_format_defaults = {"padding": 0,
+                    "usecolons": False,
                     "with_header_hide": [],
                     "without_header_hide": []}
 
@@ -32,6 +49,7 @@ _table_formats = {"simple":
                               linebetweenrows=None,
                               linebelow=Line("", "-", "  ", ""),
                               datarow=DataRow("", "  ", ""),
+                              padding=0,
                               usecolons=False,
                               with_header_hide=["linebelow"],
                               without_header_hide=[]),
@@ -39,40 +57,52 @@ _table_formats = {"simple":
                   TableFormat(None, None, None, None,
                               DataRow("", "  ", ""), **_format_defaults),
                   "grid":
-                  TableFormat(lineabove=Line("+-", "-", "-+-", "-+"),
-                              linebelowheader=Line("+=", "=", "=+=", "=+"),
-                              linebetweenrows=Line("+-", "-", "-+-", "-+"),
-                              linebelow=Line("+-", "-", "-+-", "-+"),
-                              datarow=DataRow("| ", " | ", " |"),
+                  TableFormat(lineabove=Line("+", "-", "+", "+"),
+                              linebelowheader=Line("+", "=", "+", "+"),
+                              linebetweenrows=Line("+", "-", "+", "+"),
+                              linebelow=Line("+", "-", "+", "+"),
+                              datarow=DataRow("|", "|", "|"),
+                              padding=1,
                               usecolons=False,
                               with_header_hide=[],
                               without_header_hide=["linebelowheader"]),
                   "pipe":
                   TableFormat(lineabove=None,
-                              linebelowheader=Line("| ", "-", " | ", " |"),
+                              linebelowheader=Line("|", "-", "|", "|"),
                               linebetweenrows=None,
                               linebelow=None,
-                              datarow=DataRow("| ", " | ", " |"),
+                              datarow=DataRow("|", "|", "|"),
+                              padding=1,
                               usecolons=True,
                               with_header_hide=[],
                               without_header_hide=[]),
                   "orgtbl":
                   TableFormat(lineabove=None,
-                              linebelowheader=Line("|-", "-", "-+-", "-|"),
+                              linebelowheader=Line("|", "-", "+", "|"),
                               linebetweenrows=None,
                               linebelow=None,
-                              datarow=DataRow("| ", " | ", " |"),
+                              datarow=DataRow("|", "|", "|"),
+                              padding=1,
+                              usecolons=False,
+                              with_header_hide=[],
+                              without_header_hide=["linebelowheader"]),
+                  "rst":
+                  TableFormat(lineabove=Line("", "=", "  ", ""),
+                              linebelowheader=Line("", "=", "  ", ""),
+                              linebetweenrows=None,
+                              linebelow=Line("", "=", "  ", ""),
+                              datarow=DataRow("", "  ", ""),
+                              padding=0,
                               usecolons=False,
                               with_header_hide=[],
                               without_header_hide=["linebelowheader"])}
 
-
 def simple_separated_format(separator):
     """Construct a simple TableFormat with columns separated by a separator.
 
     >>> tsv = simple_separated_format("\t") ; \
-        tabulate([["foo", 1], ["spam", 23]], tablefmt=tsv)
-    'foo \\t 1\\nspam\\t23'
+        tabulate([["foo", 1], ["spam", 23]], tablefmt=tsv) == u'foo \\t 1\\nspam\\t23'
+    True
 
     """
     return TableFormat(None, None, None, None,
@@ -85,8 +115,6 @@ def _isconvertible(conv, string):
         return True
     except ValueError:
         return False
-    except UnicodeEncodeError:
-        return False
 
 
 def _isnumber(string):
@@ -109,19 +137,20 @@ def _isint(string):
     False
     """
     return type(string) is int or \
-           isinstance(string, basestring) and  _isconvertible(int, string)
+           (isinstance(string, _binary_type) or isinstance(string, _text_type)) and \
+           _isconvertible(int, string)
 
 
 def _type(string):
-    "The least generic type (int, float, basestring, unicode)."
+    "The least generic type (int, float, str, unicode)."
     if _isint(string):
         return int
     elif _isnumber(string):
         return float
-    elif _isconvertible(str, string):
-        return basestring
+    elif isinstance(string, _binary_type):
+        return _binary_type
     else:
-        return unicode
+        return _text_type
 
 
 def _afterpoint(string):
@@ -152,24 +181,42 @@ def _afterpoint(string):
 
 
 def _padleft(width, s):
-    fmt = "{:>%ds}" % width
+    """Flush right.
+
+    >>> _padleft(6, u'\u044f\u0439\u0446\u0430') == u'  \u044f\u0439\u0446\u0430'
+    True
+
+    """
+    fmt = u"{:>%ds}" % width
     return fmt.format(s)
 
 
 def _padright(width, s):
-    fmt = "{:<%ds}" % width
+    """Flush left.
+
+    >>> _padright(6, u'\u044f\u0439\u0446\u0430') == u'\u044f\u0439\u0446\u0430  '
+    True
+
+    """
+    fmt = u"{:<%ds}" % width
     return fmt.format(s)
 
 
 def _padboth(width, s):
-    fmt = "{:^%ds}" % width
+    """Center string.
+
+    >>> _padboth(6, u'\u044f\u0439\u0446\u0430') == u' \u044f\u0439\u0446\u0430 '
+    True
+
+    """
+    fmt = u"{:^%ds}" % width
     return fmt.format(s)
 
 
 def _align_column(strings, alignment, minwidth=0):
     """[string] -> [padded_string]
 
-    >>> _align_column(["12.345", "-1234.5", "1.23", "1234.5", "1e+234", "1.0e234"], "decimal")
+    >>> list(map(str,_align_column(["12.345", "-1234.5", "1.23", "1234.5", "1e+234", "1.0e234"], "decimal")))
     ['   12.345  ', '-1234.5    ', '    1.23   ', ' 1234.5    ', '    1e+234 ', '    1.0e234']
 
     """
@@ -180,7 +227,7 @@ def _align_column(strings, alignment, minwidth=0):
         strings = [s.strip() for s in strings]
         padfn = _padboth
     elif alignment in "decimal":
-        decimals = map(_afterpoint, strings)
+        decimals = [_afterpoint(s) for s in strings]
         maxdecimals = max(decimals)
         strings = [s + (maxdecimals - decs) * " "
                    for s, decs in zip(strings, decimals)]
@@ -193,23 +240,23 @@ def _align_column(strings, alignment, minwidth=0):
 
 
 def _more_generic(type1, type2):
-    types = { int: 1, float: 2, basestring: 3, unicode: 4 }
-    invtypes = { 4: unicode, 3: basestring, 2: float, 1: int }
-    moregeneric = max(types[type1], types[type2])
+    types = { int: 1, float: 2, _text_type: 4 }
+    invtypes = { 4: _text_type, 2: float, 1: int }
+    moregeneric = max(types.get(type1, 4), types.get(type2, 4))
     return invtypes[moregeneric]
 
 
 def _column_type(strings):
     """The least generic type all column values are convertible to.
 
-    >>> _column_type(["1", "2"])
-    <type 'int'>
-    >>> _column_type(["1", "2.3"])
-    <type 'float'>
-    >>> _column_type(["1", "2.3", "four"])
-    <type 'basestring'>
-    >>> _column_type(["four", u'\u043f\u044f\u0442\u044c'])
-    <type 'unicode'>
+    >>> _column_type(["1", "2"]) is _int_type
+    True
+    >>> _column_type(["1", "2.3"]) is _float_type
+    True
+    >>> _column_type(["1", "2.3", "four"]) is _text_type
+    True
+    >>> _column_type(["four", u'\u043f\u044f\u0442\u044c']) is _text_type
+    True
 
     """
     types = map(_type, strings)
@@ -217,8 +264,19 @@ def _column_type(strings):
 
 
 def _format(val, valtype, floatfmt):
-    if valtype in [int, str, basestring]:
-        return "{}".format(val)
+    """Format a value accoding to its type.
+
+    Unicode is supported:
+
+    >>> hrow = [u'\u0431\u0443\u043a\u0432\u0430', u'\u0446\u0438\u0444\u0440\u0430'] ; \
+        tbl = [[u'\u0430\u0437', 2], [u'\u0431\u0443\u043a\u0438', 4]] ; \
+        good_result = u'\\u0431\\u0443\\u043a\\u0432\\u0430      \\u0446\\u0438\\u0444\\u0440\\u0430\\n-------  -------\\n\\u0430\\u0437             2\\n\\u0431\\u0443\\u043a\\u0438           4' ; \
+        tabulate(tbl, headers=hrow) == good_result
+    True
+
+    """
+    if valtype in [int, _binary_type, _text_type]:
+        return u"{}".format(val)
     elif valtype is float:
         return format(float(val), floatfmt)
     else:
@@ -341,10 +399,28 @@ def tabulate(list_of_lists, headers=[], tablefmt="simple",
     | spam |  41.9999 |
     | eggs | 451      |
 
+    "rst" is like a simple table format from reStructuredText; please
+    note that reStructuredText accepts also "grid" tables:
+
+    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
+    ...                ["strings", "numbers"], "rst"))
+    =========  =========
+    strings      numbers
+    =========  =========
+    spam         41.9999
+    eggs        451
+    =========  =========
+
+    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="rst"))
+    ====  ========
+    spam   41.9999
+    eggs  451
+    ====  ========
+
     """
     # format rows and columns, convert numeric values to strings
-    cols = zip(*list_of_lists)
-    coltypes = map(_column_type, cols)
+    cols = list(zip(*list_of_lists))
+    coltypes = list(map(_column_type, cols))
     cols = [[_format(v, ct, floatfmt) for v in c]
              for c,ct in zip(cols, coltypes)]
 
@@ -359,10 +435,10 @@ def tabulate(list_of_lists, headers=[], tablefmt="simple",
         minwidths = [max(minw, len(c[0])) for minw, c in zip(minwidths, cols)]
         headers = [_align_header(h, a, minw)
                    for h, a, minw in zip(headers, aligns, minwidths)]
-        rows = zip(*cols)
+        rows = list(zip(*cols))
     else:
         minwidths = [len(c[0]) for c in cols]
-        rows = zip(*cols)
+        rows = list(zip(*cols))
 
     if not isinstance(tablefmt, TableFormat):
         tablefmt = _table_formats.get(tablefmt, _table_formats["simple"])
@@ -370,25 +446,24 @@ def tabulate(list_of_lists, headers=[], tablefmt="simple",
     return _format_table(tablefmt, headers, rows, minwidths, aligns)
 
 
-def _build_row(cells, begin, sep, end):
+def _build_row(cells, padding, begin, sep, end):
     "Return a string which represents a row of data cells."
-    return (begin + sep.join(cells) + end).rstrip()
+    pad = u" "*padding
+    padded_cells = [pad + cell + pad for cell in cells]
+    return (begin + sep.join(padded_cells) + end).rstrip()
 
 
-def _build_line(colwidths, begin, fill, sep,  end):
+def _build_line(colwidths, padding, begin, fill, sep,  end):
     "Return a string which represents a horizontal line."
-    cells = [fill*w for w in colwidths]
-    return _build_row(cells, begin, sep, end)
+    cells = [fill*(w + 2*padding) for w in colwidths]
+    return _build_row(cells, 0, begin, sep, end)
 
 
 def _line_segment_with_colons(linefmt, align, colwidth):
     """Return a segment of a horizontal line with optional colons which
     indicate column's alignment (as in `pipe` output format)."""
     fill = linefmt.hline
-    # if cells are padded with spaces; .sep should be padded tool
-    # only the first symbol of .begin (and the last of .end) is used;
-    extra_width = max(0, len(linefmt.begin)+len(linefmt.end)-2)
-    w = colwidth + extra_width
+    w = colwidth
     if align in ["right", "decimal"]:
         return (fill[0] * (w - 1)) + ":"
     elif align == "center":
@@ -403,34 +478,35 @@ def _format_table(fmt, headers, rows, colwidths, colaligns):
     """Produce a plain-text representation of the table."""
     lines = []
     hidden = fmt.with_header_hide if headers else fmt.without_header_hide
+    pad = fmt.padding
 
     if fmt.lineabove and "lineabove" not in hidden:
-        lines.append(_build_line(colwidths, *fmt.lineabove))
+        lines.append(_build_line(colwidths, pad, *fmt.lineabove))
 
     if headers:
-        lines.append(_build_row(headers, *fmt.datarow))
+        lines.append(_build_row(headers, pad, *fmt.datarow))
 
     if fmt.linebelowheader and "linebelowheader" not in hidden:
         begin, fill, sep, end = fmt.linebelowheader
         if fmt.usecolons:
-            segs = [_line_segment_with_colons(fmt.linebelowheader,a,w)
+            segs = [_line_segment_with_colons(fmt.linebelowheader, a, w + 2*pad)
                     for w,a in zip(colwidths, colaligns)]
-            lines.append(_build_row(segs, begin[0], sep[len(sep)//2], end[-1]))
+            lines.append(_build_row(segs, 0, begin, sep, end))
         else:
-            lines.append(_build_line(colwidths, *fmt.linebelowheader))
+            lines.append(_build_line(colwidths, pad, *fmt.linebelowheader))
 
     if rows and fmt.linebetweenrows and "linebetweenrows" not in hidden:
         # initial rows with a line below
         for row in rows[:-1]:
-            lines.append(_build_row(row, *fmt.datarow))
-            lines.append(_build_line(colwidths, *fmt.linebetweenrows))
+            lines.append(_build_row(row, pad, *fmt.datarow))
+            lines.append(_build_line(colwidths, pad, *fmt.linebetweenrows))
         # the last row without a line below
-        lines.append(_build_row(rows[-1], *fmt.datarow))
+        lines.append(_build_row(rows[-1], pad, *fmt.datarow))
     else:
         for row in rows:
-            lines.append(_build_row(row, *fmt.datarow))
+            lines.append(_build_row(row, pad, *fmt.datarow))
 
     if fmt.linebelow and "linebelow" not in hidden:
-        lines.append(_build_line(colwidths, *fmt.linebelow))
+        lines.append(_build_line(colwidths, pad, *fmt.linebelow))
 
     return "\n".join(lines)
